#include <X11/Xlib.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <glib.h>
#include <glib/gprintf.h>

#define DBUS_API_SUBJECT_TO_CHANGE
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <locale.h>
#include <string.h>
#include <time.h>

//#define DEBUG
//#define VERBOSE

#include "config.h"
#include "dbg.h"
#include "bg.h"
#include "gtkbgbox.h"
#include "sexy-url-label.h"
#include "infoboxd.h"

/**********************************************************************/
/*                widget callbacks functions                 */
static gboolean ib_gui_callback_timeout(gpointer data)
{
    ENTER;
    if(ib.aftertimeout==TIMEOUT_BELOW)
        ib_gui_notify_hide(WINDOW_BELOW, NOTIFY_EXPIRED);
    else if(ib.aftertimeout==TIMEOUT_HIDE)
        ib_gui_notify_hide(WINDOW_HIDE, NOTIFY_EXPIRED);

    RET(FALSE);
}

static gboolean ib_gui_callback_expose_event(GtkWidget *widget, GdkEventExpose *e, gpointer data)
{
    ENTER;
    GdkGC *gc;
    GdkColormap *gcm;

    gcm=gtk_widget_get_colormap(widget);
    gc=gdk_gc_new(GDK_DRAWABLE(ib.gwin->window));
    gdk_gc_set_rgb_fg_color(gc, &ib.bordercolor);
    gdk_draw_rectangle(GDK_DRAWABLE(ib.gwin->window), gc, TRUE, 0, 0, ib.cw, ib.ch);

    RET(FALSE);
}

static gboolean ib_gui_callback_button_release_event(GtkWidget *widget, GdkEventButton *e, gpointer data)
{
    ENTER;
    if(!ib.urlclicked)
    {
        if(e->button==1 && ib.windowstate!=WINDOW_ABOVE)
        {
            gdk_window_set_keep_above(ib.gwin->window, TRUE);
            ib.windowstate=WINDOW_ABOVE;
        }
        else if(e->button==1 && ib.windowstate==WINDOW_ABOVE && ib.notifyhided==FALSE)
        {
            ib_dbus_hide_notify(NOTIFY_DISSMISED);
            g_source_remove(ib.timeoutid);
            ib.notifyhided=TRUE;
            if(ib.transparncy)
                gtk_bgbox_set_background(ib.bbox, BG_ROOT, ib.tintcolor, ib.alpha);
        }
        else if(e->button==2)
            ib_gui_notify_hide(WINDOW_HIDE, NOTIFY_DISSMISED);
        else if(e->button==3 && ib.windowstate!=WINDOW_BELOW)
            ib_gui_notify_hide(WINDOW_BELOW, NOTIFY_DISSMISED);
    }
    ib.urlclicked=FALSE;

    RET(TRUE);
}

/*static gboolean ib_gui_callback_scroll_event(GtkWidget *widget, GdkEventScroll *e, gpointer data)
{
    ENTER;
	gchar *msg;

	msg=g_strdup_printf("rolnołeś mnie w %d stronę!", e->direction);
	//ib_set_infobox_notify_msg("InfoBox", msg);
	gtk_widget_hide(ib.image);
	ib_gui_notify_show(NULL, msg, NULL, NULL, 0);

	g_free(msg);
	RET(TRUE);
}*/

static gboolean ib_gui_callback_configure_event (GtkWidget *widget, GdkEventConfigure *e, gpointer data)
{
    ENTER;

    if(e->width==ib.cw && e->height==ib.ch && e->x==ib.cx && e->y==ib.cy)
        RET(FALSE);

    ib.cw=e->width;
    ib.ch=e->height;
    ib.cx=e->x;
    ib.cy=e->y;
    DBG("New width (%d) and new height (%d)\n", ib.cw, ib.ch);

    if(ib.transparncy)
        fb_bg_notify_changed_bg(ib.bg);

	RET(FALSE);
}

static gboolean ib_gui_callback_size_req(GtkWidget *widget, GtkRequisition *r, gpointer data)
{
    ENTER;

    ib_gui_calculate_width(r->width);
    ib_gui_calculate_height(r->height);
    r->width=ib.cw;
    r->height=ib.ch;

    RET(FALSE);
}

static void ib_gui_callback_size_alloc(GtkWidget *widget, GtkAllocation *a, gpointer data)
{
    ENTER;
    gint scrw, scrh;

    scrw=gdk_screen_width();
    scrh=gdk_screen_height();

    if(ib.horiz==HORIZ_NONE)
        a->x=ib.geomx+ib.xoffset;
    else if(ib.horiz==HORIZ_LEFT)
        a->x=ib.xoffset;
    else if(ib.horiz==HORIZ_CENTER)
        a->x=scrw/2-ib.cw/2+ib.xoffset;
    else //if(ib.horiz==HORIZ_RIGHT)
        a->x=scrw-ib.cw+ib.xoffset;

    if(ib.vert==VERT_NONE)
        a->y=ib.geomy+ib.yoffset;
    else if(ib.vert==VERT_TOP)
        a->y=ib.yoffset;
    else if(ib.vert==VERT_MIDDLE)
        a->y=scrh/2-ib.ch/2+ib.yoffset;
    else //if(ib.vert==VERT_BOTTOM)
        a->y=scrh-ib.ch+ib.yoffset;

    DBG("We're moving window to (%d,%d)\n", a->x, a->y);
    gtk_window_move(GTK_WINDOW(ib.gwin), a->x, a->y);

    RET();
}

static gboolean ib_gui_callback_image_size_req(GtkWidget *widget, GtkRequisition *req, gpointer data)
{
    ENTER;
	if(ib.imagewmax>=0 && req->width>ib.imagewmax)
		req->width=ib.imagewmax;
    
	if(ib.imagehmax>=0 && req->height>ib.imagehmax)
        req->height=ib.imagehmax;

	RET(TRUE);
}

static void ib_gui_callback_url_activated(GtkWidget *url_label, const gchar *url)
{
    ENTER;
	gchar *escaped_url, *cmd;

	escaped_url = g_shell_quote(url);
	cmd = g_strdup_printf("gnome-open %s", escaped_url);
	DBG("user clicked a link %s\n", escaped_url);
	g_spawn_command_line_async(cmd, NULL);
    ib.urlclicked=TRUE;

	g_free(escaped_url);
	g_free(cmd);

    RET();
}
/*====================================================================*/


/**********************************************************************/
/*                        config funtions                        */
static gint ib_config_strcmp(gchar *string1, gchar *string2)
{
    gchar *str1, *str2;
    gint ret;

    str1=g_utf8_casefold(string1, -1);
    str2=g_utf8_casefold(string2, -1);

    ret=g_utf8_collate(str1, str2);

    g_free(str1);
    g_free(str2);
    return(ret);
}

static void ib_config_parser(GKeyFile *keyfile, gchar *group, gchar **keys)
{
    ENTER;
    register gint i, j;
    gchar *val;
    GError *error=NULL;

    for(i=0; keys[i]; ++i)
	{
		val=NULL;
		for(j=0; j<sizeof(configtable)/sizeof(cfg); j++)
		{
			if(ib_config_strcmp(configtable[j].key, keys[i])==0)
			{
				switch(configtable[j].type)
				{
					case CFG_INTEGER:
                        val=g_key_file_get_string(keyfile, group, keys[i], NULL);
                        sscanf(val, "%d", (gint *) configtable[j].val);

                        g_free(val);
						break;
					case CFG_BOOLEAN:
                    {
                        *configtable[j].val=g_key_file_get_boolean(keyfile, group, keys[i], &error);
						break;
                    }
					case CFG_STRING:
                        configtable[j].val=g_key_file_get_string(keyfile, group, keys[i], NULL);
						break;
                    case CFG_TINTCOLOR:
                    {
                        GdkColor color;
                        guint32 r, g, b, *p;
                        val=g_key_file_get_string(keyfile, group, keys[i], NULL);
                        gdk_color_parse(val, &color);
                        r=((gfloat) color.red*255.0/65535.0);
                        g=((gfloat) color.green*255.0/65535.0);
                        b=((gfloat) color.blue*255.0/65535.0);
                        p=(guint32 *) configtable[j].val;
                        *p=(0<<24)+(r<<16)+(g<<8)+b;

                        g_free(val);
                        break;
                    }
                    case CFG_SIZETYPE:
                    {
                        val=g_key_file_get_string(keyfile, group, keys[i], NULL);
                        if(ib_config_strcmp(val, "none")==0)
                            *configtable[j].val=SIZE_NONE;
                        else if(ib_config_strcmp(val, "pixel")==0)
                            *configtable[j].val=SIZE_PIXEL;
                        else if(ib_config_strcmp(val, "percent")==0)
                            *configtable[j].val=SIZE_PERCENT;

                        g_free(val);
                        break;
                    }
                    case CFG_JUSTIFY:
                    {
                        val=g_key_file_get_string(keyfile, group, keys[i], NULL);
                        if(ib_config_strcmp(val, "left")==0)
                            *configtable[j].val=JUSTIFY_LEFT;
                        else if(ib_config_strcmp(val, "right")==0)
                            *configtable[j].val=JUSTIFY_RIGHT;
                        else if(ib_config_strcmp(val, "center")==0)
                            *configtable[j].val=JUSTIFY_CENTER;

                        g_free(val);
                        break;
                    }
                    case CFG_FONT:
                    {
                        PangoFontDescription *font;
                        gchar **f;
                        val=g_key_file_get_string(keyfile, group, keys[i], NULL);
                        pango_font_description_free(ib.textfont);
                        font=pango_font_description_from_string(val);
                        f=(gchar **) configtable[j].val;
                        *f=(gchar *) font;

                        g_free(val);
                        break;
                    }
                    case CFG_GDKCOLOR:
                    {
                        val=g_key_file_get_string(keyfile, group, keys[i], NULL);
                        gdk_color_parse(val, (GdkColor *) configtable[j].val);

                        g_free(val);
                        break;
                    }
                    case CFG_HORIZ:
                    {
                        val=g_key_file_get_string(keyfile, group, keys[i], NULL);
                        if(ib_config_strcmp(val, "none")==0)
                            *configtable[j].val=HORIZ_NONE;
                        else if(ib_config_strcmp(val, "left")==0)
                            *configtable[j].val=HORIZ_LEFT;
                        else if(ib_config_strcmp(val, "right")==0)
                            *configtable[j].val=HORIZ_RIGHT;
                        else if(ib_config_strcmp(val, "center")==0)
                            *configtable[j].val=HORIZ_CENTER;

                        g_free(val);
                        break;
                    }
                    case CFG_VERT:
                    {
                        val=g_key_file_get_string(keyfile, group, keys[i], NULL);
                        if(ib_config_strcmp(val, "none")==0)
                            *configtable[j].val=VERT_NONE;
                        else if(ib_config_strcmp(val, "top")==0)
                            *configtable[j].val=VERT_TOP;
                        else if(ib_config_strcmp(val, "middle")==0)
                            *configtable[j].val=VERT_MIDDLE;
                        else if(ib_config_strcmp(val, "bottom")==0)
                            *configtable[j].val=VERT_BOTTOM;

                        g_free(val);
                        break;
                    }
                    case CFG_AFTERTIMEOUT:
                    {
                        val=g_key_file_get_string(keyfile, group, keys[i], NULL);
                        if(ib_config_strcmp(val, "below")==0)
                            *configtable[j].val=TIMEOUT_BELOW;
                        else if(ib_config_strcmp(val, "hide")==0)
                            *configtable[j].val=TIMEOUT_HIDE;

                        g_free(val);
                        break;
                    }
				}
				break;
			}
		}
        if(j==sizeof(configtable)/sizeof(cfg))
            DBG("Unknown element: %s\n", keys[i]);
	}

    RET();
}

static void ib_config_from_file(void)
{
    ENTER;
	GKeyFile *keyfile;
    GError *error=NULL;
    gchar *configfile, *startgroup, **keys;
    gint len;

    configfile=g_strdup_printf("%s/.infobox/infoboxd.cfg", g_get_home_dir());
	keyfile=g_key_file_new();
	if(!g_key_file_load_from_file(keyfile, configfile, G_KEY_FILE_NONE, &error))
    {
        DBG("Couldn't load config file: %s\n", error->message);
        g_error_free(error);
        return;
    }
	startgroup=g_key_file_get_start_group(keyfile);
    if(ib_config_strcmp(startgroup, "InfoBoxDaemon")!=0)
    {
        DBG("Error in config file: wrong group name, should be \"InfoBoxd\"\n");
        return;
    }

	keys=g_key_file_get_keys(keyfile, startgroup, &len, NULL);
    ib_config_parser(keyfile, startgroup, keys);

	g_strfreev(keys);
	g_free(startgroup);

	g_key_file_free(keyfile);
    g_free(configfile);
    RET();
}

static void ib_config_setup_gui(void)
{
    ENTER;
    // we freeze window to not flicker
    gdk_window_freeze_updates(ib.gwin->window);

    if(ib.transparncy)
    {
        gtk_rc_parse_string(transparent_rc);
        ib.bg=fb_bg_get_for_display();
        if(ib.notifyhided)
            gtk_bgbox_set_background(ib.bbox, BG_ROOT, ib.tintcolor, ib.alpha);
        else
            gtk_bgbox_set_background(ib.bbox, BG_ROOT, ib.hilighttintcolor, ib.alpha);
    }
    else
    {
        gtk_bgbox_set_background(ib.bbox, BG_STYLE, 0, 0);
    }

    // setting border (color is set by parser)
    gtk_container_set_border_width(GTK_CONTAINER(ib.gwin), ib.borderwidth);

	// setting time label
    gtk_label_set_justify(GTK_LABEL(ib.time), GTK_JUSTIFY_RIGHT);
    gtk_widget_modify_fg(ib.time, GTK_STATE_NORMAL, &ib.textcolor);
    gtk_widget_modify_font(ib.time, ib.textfont);

	// setting title label
    gtk_label_set_justify(GTK_LABEL(ib.title), ib.textjustify);
    gtk_widget_modify_fg(ib.title, GTK_STATE_NORMAL, &ib.textcolor);
    gtk_widget_modify_font(ib.title, ib.textfont);

    // setting image
    gtk_image_set_pixel_size(GTK_IMAGE(ib.image), ib.iconsize);

	// setting body label
    gtk_label_set_justify(GTK_LABEL(ib.body), ib.textjustify);
    gtk_widget_modify_fg(ib.body, GTK_STATE_NORMAL, &ib.textcolor);
    gtk_widget_modify_font(ib.body, ib.textfont);

    // now we can update window
    gdk_window_thaw_updates(ib.gwin->window);
    gtk_widget_queue_draw_area(ib.gwin, 0, 0, ib.cw, ib.ch);

    RET();
}

static void ib_config_set(void)
{
    ENTER;
	ib.geomx=200;
	ib.geomy=400;
	ib.horiz=HORIZ_NONE;
    ib.vert=VERT_NONE;
	ib.textjustify=JUSTIFY_LEFT;
    gdk_color_parse("#FFFFFF", &ib.textcolor);
    ib.textfont=pango_font_description_from_string("");
	ib.widthtype=SIZE_NONE;
	ib.width=100;
	ib.widthmin=-1;
	ib.widthmax=-1;
	ib.heighttype=SIZE_NONE;
	ib.height=100;
	ib.heightmin=-1;
	ib.heightmax=-1;
	ib.xoffset=0;
	ib.yoffset=0;
    ib.transparncy=TRUE;
	ib.alpha=128;
    ib.tintcolor=0x00000000;
    ib.hilighttintcolor=0x00000066;
    ib.borderwidth=1;
    gdk_color_parse("#777777", &ib.bordercolor);
    ib.timeout=15;
    ib.aftertimeout=TIMEOUT_BELOW;
    ib.iconsize=48;
    ib.imagewmax=64;
    ib.imagehmax=64;
    ib.notifyid=0;
    ib.notifyhided=TRUE;

    ib_config_from_file();
   
    if(ib.width<0)
        ib.width=100;
    if(ib.widthtype==SIZE_PERCENT && ib.width>100)
        ib.width=100;

    if(ib.heighttype==SIZE_PIXEL)
        ib.height=CLAMP(ib.height, ib.heightmin, ib.heightmax);

    ib.alpha=CLAMP(ib.alpha, 0, 255);

    RET();
}
/*====================================================================*/


/**********************************************************************/
/*                        gui control funtions                        */
static gint ib_gui_calculate_width(gint width)
{
    ENTER;
    gint scrw;

    scrw=gdk_screen_width();
    if(ib.widthtype==SIZE_NONE)
    {
        if(ib.widthmin>-1 && width<ib.widthmin)
        {
            gtk_widget_set_size_request(ib.gwin, ib.widthmin, -1);
            ib.cw=ib.widthmin;
        }
        else if(ib.widthmax>-1 && width>ib.widthmax)
        {
            gtk_widget_set_size_request(ib.gwin, ib.widthmax, -1);
            ib.cw=ib.widthmax;
        }
        else
            ib.cw=width;
    }
    else if(ib.widthtype==SIZE_PIXEL)
        ib.cw=ib.width;
    else if(ib.widthtype==SIZE_PERCENT)
        ib.cw=(ib.width*scrw)/100;
    DBG("width=%d\n", ib.cw);

    RET(FALSE);
}

static gint ib_gui_calculate_height(gint height)
{
    ENTER;
    gint scrh;

    scrh=gdk_screen_height();
    if(ib.heighttype==SIZE_NONE)
    {
        if(ib.heightmin>-1 && height<ib.heightmin)
        {
            gtk_widget_set_size_request(ib.gwin, -1, ib.heightmin);
            ib.ch=ib.heightmin;
        }
        else if(ib.heightmax>-1 && height>ib.heightmax)
        {
            gtk_widget_set_size_request(ib.gwin, -1, ib.heightmax);
            ib.ch=ib.heightmax;
        }
        else
            ib.ch=height;
    }
    else if(ib.heighttype==SIZE_PIXEL)
        ib.ch=ib.height;
    else if(ib.heighttype==SIZE_PERCENT)
        ib.ch=(ib.height*scrh)/100;
    DBG("height=%d\n", ib.ch);

    RET(FALSE);
}

static GdkPixbuf *ib_gui_load_image(guchar *data, gboolean raw, gint len, gint scaledwidth, gint scaledheight)
{
    ENTER;
    GdkPixbuf *pix=NULL;
    GError *error=NULL;

    // raw image data
    if(raw==TRUE)
    {
        GdkPixbufLoader *loader;

        loader=gdk_pixbuf_loader_new();
        gdk_pixbuf_loader_write(loader, data, len, &error);
		if(error)
        {
			g_object_unref(loader);
            DBG("loading image from raw data error: %s\n", error->message);
            g_error_free(error);
			RET(NULL);
		}

        pix=gdk_pixbuf_loader_get_pixbuf(loader);
		if(!pix)
        {
            DBG("Could not get pixbuf from loader\n");
            RET(0);
        }
        g_object_ref(G_OBJECT(pix));

        gdk_pixbuf_loader_close(loader, NULL);
        g_object_unref(loader);
        DBG("Loaded raw image\n");
    }
    // absolute path to image
    else if((*data)=='/')
    {
        pix=(GdkPixbuf *) gdk_pixbuf_animation_new_from_file(data, &error);
        if(error)
        {
            DBG("Could not load file: %s\n", error->message);
            g_error_free(error);
            RET(NULL);
        }
        if(gdk_pixbuf_animation_is_static_image(GDK_PIXBUF_ANIMATION(pix)))
        {
            g_object_unref(G_OBJECT(pix));
            pix=gdk_pixbuf_new_from_file(data, &error);
            if(error)
            {
                DBG("Could not load file: %s\n", error->message);
                g_error_free(error);
                RET(NULL);
            }
        }

        DBG("Loaded image from file\n");
        if(pix)
            g_object_ref(G_OBJECT(pix)); // 'cause we will later unref after showing a notify
    }
    // theme icon
    else
    {
        GtkIconTheme *icon_theme;

        icon_theme=gtk_icon_theme_get_default();
        pix=gtk_icon_theme_load_icon(icon_theme, data, ib.iconsize, GTK_ICON_LOOKUP_USE_BUILTIN, &error);
        if(!pix)
        {
            DBG("Couldn't load icon: %s\n", error->message);
            g_error_free(error);
            RET(NULL);
        }
        DBG("Loaded theme icon\n");
    }

    if(pix)
    {
        gfloat width, height, factor, x, y;

        if(GDK_IS_PIXBUF_ANIMATION(pix))
        {
            width=gdk_pixbuf_animation_get_width(GDK_PIXBUF_ANIMATION(pix));
            height=gdk_pixbuf_animation_get_height(GDK_PIXBUF_ANIMATION(pix));
        }
        else
        {
            width=gdk_pixbuf_get_width(pix);
            height=gdk_pixbuf_get_height(pix);
        }

        DBG("image's width=%f, height=%f, scaledwidth=%d, scaledheight=%d\n", width, height, scaledwidth, scaledheight);
        if(width>scaledwidth || height>scaledheight)
        {
            GdkPixbuf *newpix;

            if(GDK_IS_PIXBUF_ANIMATION(pix))
            {
                GdkPixbufAnimation *anim;
    
                DBG("it's a too big animation, so we will makit static to scall it\n");
                anim=GDK_PIXBUF_ANIMATION(pix);
                pix=gdk_pixbuf_animation_get_static_image(anim);
                g_object_unref(G_OBJECT(anim));
    
                if((*data)=='/')
                    g_object_ref(G_OBJECT(pix)); // 'cause we will later unref after showing a notify
            }

        x=ABS(scaledwidth-width);
            y=ABS(scaledheight-height);
            if(scaledwidth>0 && (x>=y || (x>0 && scaledheight==-1)))
            {
                factor=width/height;
                height=scaledwidth/factor;
                width=scaledwidth;
            }
            else if(scaledheight>0 && (y>x || (y>0 && scaledwidth==-1)))
            {
                factor=height/width;
                width=scaledheight/factor;
                height=scaledheight;
            }
            else
                RET(pix);

            DBG("image's rescaled width=%d, height=%d\n", (gint) width, (gint) height);
            newpix=gdk_pixbuf_scale_simple(pix, (gint) width, (gint) height, GDK_INTERP_BILINEAR);
            g_object_unref(pix);
            if(raw || (*data)=='/')
            {
                g_object_ref(G_OBJECT(newpix));
                g_object_unref(pix);
            }
            pix=newpix;
        }
    }

    RET(pix);
}

static void ib_gui_notify_hide(gint how, gint reason)
{
    ENTER;
    if(how==WINDOW_BELOW && ib.windowstate==WINDOW_ABOVE)
    {
        gdk_window_set_keep_below(ib.gwin->window, TRUE);
        ib.windowstate=WINDOW_BELOW;
    }
    else if(how==WINDOW_HIDE) 
    {
        gtk_widget_hide(ib.gwin);
        ib.windowstate=WINDOW_HIDE;
    }

    if(ib.notifyhided==FALSE)
    {
        if(ib.transparncy)
            gtk_bgbox_set_background(ib.bbox, BG_ROOT, ib.tintcolor, ib.alpha);
        ib_dbus_hide_notify(reason);
        g_source_remove(ib.timeoutid);
        ib.notifyhided=TRUE;
    } 

    RET();
}

static void ib_gui_notify_show(GdkPixbuf *icon, gchar *title, gchar* body, GdkPixbufAnimation *image, gint timeout)
{
	ENTER;
	GtkRequisition req;
	struct tm tm;
	time_t tp;
	gchar locale_time[14], *format;
    gint iconwidth=0, imagewidth=0, timewidth, sum;

    gdk_window_freeze_updates(ib.gwin->window);

	time(&tp);
	localtime_r(&tp, &tm);

    if(!icon)
        format="[%H:%M:%S] ";
    else
        format="[%H:%M:%S]";
	strftime(locale_time, 13, format, &tm);
    gtk_label_set_markup(GTK_LABEL(ib.time), locale_time);

    if(GDK_IS_PIXBUF_ANIMATION(icon))
        gtk_image_set_from_animation(GTK_IMAGE(ib.icon), GDK_PIXBUF_ANIMATION(icon));
    else
        gtk_image_set_from_pixbuf(GTK_IMAGE(ib.icon), icon);
    if(icon)
    {
        gtk_widget_show(ib.icon);
        gtk_widget_size_request(ib.title, &req);
        gtk_widget_set_size_request(ib.icon, req.height, req.height);
    }
    else
        gtk_widget_hide(ib.icon);

    if(ib.textjustify==JUSTIFY_LEFT || (ib.widthtype==SIZE_NONE && ib.widthmax==-1))
        gtk_label_set_line_wrap(GTK_LABEL(ib.title), FALSE);
    else
        gtk_label_set_line_wrap(GTK_LABEL(ib.title), TRUE);
    sexy_url_label_set_markup(SEXY_URL_LABEL(ib.title), title);

    if(GDK_IS_PIXBUF_ANIMATION(image))
        gtk_image_set_from_animation(GTK_IMAGE(ib.image), image);
    else
        gtk_image_set_from_pixbuf(GTK_IMAGE(ib.image), GDK_PIXBUF(image));
    gtk_widget_show(ib.image);

	if(!body || !*body)
    {
        if(!image)
            gtk_widget_hide(ib.image);
		gtk_widget_hide(ib.body);
    }
	else
	{
		gtk_widget_show(ib.body);
		sexy_url_label_set_markup(SEXY_URL_LABEL(ib.body), body);
	}

// reset sizes of time, icon, title, image and body
// so GTK+ will set the best it choosed
    gtk_widget_set_size_request(ib.time, -1, -1);
    gtk_widget_set_size_request(ib.icon, -1, -1);
    gtk_widget_set_size_request(ib.title, -1, -1);
    gtk_widget_set_size_request(ib.image, -1, -1);
    gtk_widget_set_size_request(ib.body, -1, -1);

    if(icon)
    {
        gtk_widget_size_request(ib.icon, &req);
        iconwidth=req.width+2;
    }

// waiting for infobox client to test it 
    gtk_widget_size_request(ib.image, &req);
    imagewidth=req.width;
    gtk_widget_size_request(ib.time, &req);
    timewidth=req.width;
    sum=timewidth+iconwidth+2;
    DBG("iconwidth=%d; imagewidth=%d; timewidth=%d; sum=%d\n", iconwidth, imagewidth, timewidth, sum);
    if(image || body)
    {
        if(sum>imagewidth)
            gtk_widget_set_size_request(ib.image, sum, -1);
        else if(sum<imagewidth)
        {
            gtk_widget_set_size_request(ib.time, imagewidth-iconwidth, -1);
            gtk_misc_set_alignment(GTK_MISC(ib.time), imagewidth-iconwidth-timewidth, 0);
        }

        gtk_widget_size_request(ib.gwin, &req); // we need this, cause configure_event is not sent
        ib.cw=req.width;
        gtk_widget_set_size_request(ib.body, ib.cw-sum-4-2*ib.borderwidth, -1);
    }

    gtk_widget_set_size_request(ib.title, ib.cw-sum-4-2*ib.borderwidth, -1);

    if(ib.transparncy)
        gtk_bgbox_set_background(ib.bbox, BG_ROOT, ib.hilighttintcolor, ib.alpha);

    gdk_window_thaw_updates(ib.gwin->window);
    gtk_widget_show(ib.gwin);

    if(ib.notifyhided==FALSE)
    {
        ib_dbus_hide_notify(NOTIFY_EXPIRED);
        g_source_remove(ib.timeoutid);        
    }
    ib.notifyhided=FALSE;
    ib.urlclicked=FALSE;
    ib.windowstate=WINDOW_ABOVE;
    ib.timeoutid=g_timeout_add(ib.timeout*1000, ib_gui_callback_timeout, NULL);
    gdk_window_set_keep_above(ib.gwin->window, TRUE);

    RET();
}

static void ib_gui_start(void)
{
    ENTER;
	ib.gwin=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(ib.gwin), "InfoBox");
    gtk_window_set_wmclass(GTK_WINDOW(ib.gwin), "info-box-daemon-notify", "info-box-daemon-notify");
    gtk_container_set_border_width(GTK_CONTAINER(ib.gwin), ib.borderwidth);
    gtk_window_set_resizable(GTK_WINDOW(ib.gwin), FALSE);
    gtk_window_set_position(GTK_WINDOW(ib.gwin), GTK_WIN_POS_NONE);
    gtk_window_set_decorated(GTK_WINDOW(ib.gwin), FALSE);
    gtk_window_stick(GTK_WINDOW(ib.gwin));
    gtk_window_set_type_hint(GTK_WINDOW(ib.gwin), GDK_WINDOW_TYPE_HINT_DOCK);
    gtk_window_set_skip_taskbar_hint(GTK_WINDOW(ib.gwin), TRUE);
    gtk_window_set_skip_pager_hint(GTK_WINDOW(ib.gwin), TRUE);
    gtk_window_set_accept_focus(GTK_WINDOW(ib.gwin), FALSE);
    ib.windowstate=WINDOW_HIDE;
    
	g_signal_connect(G_OBJECT(ib.gwin), "destroy", G_CALLBACK(gtk_main_quit), NULL);
	g_signal_connect(G_OBJECT(ib.gwin), "configure-event", G_CALLBACK(ib_gui_callback_configure_event), NULL);
    g_signal_connect(G_OBJECT(ib.gwin), "expose-event", G_CALLBACK(ib_gui_callback_expose_event), NULL);
	g_signal_connect(G_OBJECT(ib.gwin), "size-request",	G_CALLBACK(ib_gui_callback_size_req), NULL);
	g_signal_connect(G_OBJECT(ib.gwin), "size-allocate", G_CALLBACK(ib_gui_callback_size_alloc), NULL);

    gtk_widget_realize(ib.gwin);
    gtk_widget_set_app_paintable(ib.gwin, TRUE);

	ib.bbox=gtk_bgbox_new();
    gtk_widget_show(ib.bbox);
    gtk_container_add(GTK_CONTAINER(ib.gwin), ib.bbox);
    gtk_container_set_border_width(GTK_CONTAINER(ib.bbox), 2);
    if(ib.transparncy)
    {
        gtk_rc_parse_string(transparent_rc);
        ib.bg=fb_bg_get_for_display();
        gtk_bgbox_set_background(ib.bbox, BG_ROOT, ib.tintcolor, ib.alpha);
    }

	ib.ebox=gtk_event_box_new();
    gtk_widget_show(ib.ebox);
	gtk_container_add(GTK_CONTAINER(ib.bbox), ib.ebox);
    g_signal_connect(G_OBJECT(ib.ebox), "button-release-event",
          G_CALLBACK(ib_gui_callback_button_release_event), NULL);
//    g_signal_connect(G_OBJECT(ib.ebox), "scroll-event",
//          G_CALLBACK(ib_gui_callback_scroll_event), NULL);

	ib.vbox=gtk_vbox_new(FALSE, 0);
    gtk_widget_show(ib.vbox);
	gtk_container_add(GTK_CONTAINER(ib.ebox), ib.vbox);

	ib.h1box=gtk_hbox_new(FALSE, 2);
    gtk_widget_show(ib.h1box);
	gtk_box_pack_start(GTK_BOX(ib.vbox), ib.h1box, TRUE, TRUE, 0);
	ib.h2box=gtk_hbox_new(FALSE, 0);
    gtk_widget_show(ib.h2box);
	gtk_box_pack_start(GTK_BOX(ib.vbox), ib.h2box, TRUE, TRUE, 0);

	ib.time=gtk_label_new(NULL);
    gtk_widget_show(ib.time);
    gtk_misc_set_alignment(GTK_MISC(ib.time), 0, 0);
	gtk_box_pack_start(GTK_BOX(ib.h1box), ib.time, FALSE, TRUE, 0);
    
    ib.icon=gtk_image_new_from_pixbuf(NULL);
    gtk_widget_show(ib.icon);
	gtk_box_pack_start(GTK_BOX(ib.h1box), ib.icon, FALSE, FALSE, 0);

	ib.title=sexy_url_label_new();
    gtk_widget_show(ib.title);
	gtk_box_pack_start(GTK_BOX(ib.h1box), ib.title, TRUE, TRUE, 0);
    gtk_misc_set_alignment(GTK_MISC(ib.title), 0, 0);
	gtk_label_set_line_wrap(GTK_LABEL(ib.title), TRUE);
    g_signal_connect(G_OBJECT(ib.title), "url_activated", G_CALLBACK(ib_gui_callback_url_activated), NULL);

	ib.vimgbox=gtk_vbox_new(FALSE, 0);
    gtk_widget_show(ib.vimgbox);
	gtk_box_pack_start(GTK_BOX(ib.h2box), ib.vimgbox, FALSE, FALSE, 0);

	ib.image=gtk_image_new_from_pixbuf(NULL);
    gtk_widget_show(ib.image);
    g_signal_connect(G_OBJECT(ib.image), "size-request", G_CALLBACK(ib_gui_callback_image_size_req), NULL);
	gtk_box_pack_start(GTK_BOX(ib.vimgbox), ib.image, FALSE, FALSE, 0);

	ib.body=sexy_url_label_new();
    gtk_widget_show(ib.body);
	gtk_box_pack_start(GTK_BOX(ib.h2box), ib.body, TRUE, TRUE, 0);
	gtk_misc_set_alignment(GTK_MISC(ib.body), 0, 0);
	gtk_label_set_line_wrap(GTK_LABEL(ib.body), TRUE);
    g_signal_connect(G_OBJECT(ib.body), "url_activated", G_CALLBACK(ib_gui_callback_url_activated), NULL);

    RET();
}
/*====================================================================*/


/**********************************************************************/
/*                            dbus control                            */

gchar *ib_dbus_extract_image_from_array(DBusMessageIter *msgiter)
{
    ENTER;
    DBusMessageIter iter, sub;
    gint arraytype;
    guchar *data=NULL, *tmp=NULL;
    gint len=0, tmplen, imgs=0;

    dbus_message_iter_recurse(msgiter, &iter);

    arraytype=dbus_message_iter_get_element_type(msgiter);
    if(arraytype==DBUS_TYPE_STRING || arraytype==DBUS_TYPE_ARRAY)
    {
        if(dbus_message_iter_get_arg_type(&iter)!=DBUS_TYPE_INVALID)
        {
            do
            {
                imgs++;
                if(arraytype==DBUS_TYPE_STRING)
                {
                    dbus_message_iter_get_basic(&iter, &data);
                    DBG("file: %s\n", data);

                    // we will use only one frame
                    break;
                }
                else if(arraytype==DBUS_TYPE_ARRAY)
                {
                    dbus_message_iter_recurse(&iter, &sub);
                    dbus_message_iter_get_fixed_array(&sub, &tmp, &tmplen);
                    data=g_realloc(data, tmplen);
                    g_memmove(data+len, tmp, tmplen);
                    len+=tmplen;
                }
            } while(dbus_message_iter_next(&iter));
        }

        DBG("There are %d images\n", imgs);
    }

    RET(data);
}

static void ib_dbus_hide_notify(gint reason)
{
    ENTER;
    if(ib.notifyid)
    {
        DBusMessage *signal;
        DBusMessageIter iter;
    
        signal=dbus_message_new_signal("/org/freedesktop/Notifications", "org.freedesktop.Notifications", "NotificationClosed");
        DBG("Sending NotificationClosed signal with id %d\n", ib.notifyid);
        dbus_message_iter_init_append(signal, &iter);
        dbus_message_iter_append_basic(&iter, DBUS_TYPE_UINT32, &ib.notifyid);
        dbus_message_iter_append_basic(&iter, DBUS_TYPE_UINT32, &reason);
        dbus_connection_send(ib.dbusconn, signal, NULL);
    
        dbus_message_unref(signal);
    }

    RET();
}

static DBusMessage *ib_dbus_handle_notify(DBusMessage *message, gboolean shortmsg)
{
    ENTER;
    DBusMessageIter iter;
    DBusMessage *reply=NULL;
    GdkPixbufAnimation *pix=NULL;
    GdkPixbuf *icon=NULL;
    gint type, id;
    gchar *title=NULL;

    // args
    gchar *app_name=NULL;
    gchar *app_icon=NULL;
    guint32 replaces_id=0;
    gchar *notification_type=NULL;
    gchar urgency_level=0;
    gchar *summary=NULL;
    gchar *body=NULL;
    gchar *images=NULL;
    gboolean expires=TRUE;
    guint expire_timeout=ib.timeout;

    dbus_message_iter_init(message, &iter);

    // App name
    type=dbus_message_iter_get_arg_type(&iter);
    if(type!=DBUS_TYPE_STRING)
    {
        DBG("Invalid notify message. app name argument is not a string\n");
        RET(0);
    }
    dbus_message_iter_get_basic(&iter, &app_name);
    DBG("app_name; '%s'\n", app_name);
    dbus_message_iter_next(&iter);

    // App icon -- we don't care
    type=dbus_message_iter_get_arg_type(&iter);
    if(type==DBUS_TYPE_STRING)
    {
        dbus_message_iter_get_basic(&iter, &app_icon);
        DBG("app_icon; '%s'\n", app_icon);
    }
    else if(type==DBUS_TYPE_ARRAY)
    {
        app_icon=ib_dbus_extract_image_from_array(&iter);
    }
    else
    {
        DBG("Invalid notify message. app icon argument is not an array or a string\n");
        RET(0);
    }
    dbus_message_iter_next(&iter);

    if(shortmsg==FALSE)
    {
        // Replaces ID
        type=dbus_message_iter_get_arg_type(&iter);
        if(type!=DBUS_TYPE_UINT32)
        {
            DBG("Invalid notify message. Expires replaces id is not uint32\n");
            RET(0);
        }
        dbus_message_iter_get_basic(&iter, &replaces_id);
        DBG("replaces_id; %d\n", replaces_id);
        dbus_message_iter_next(&iter);
        
        // Notification type -- not yet implempeted
        type=dbus_message_iter_get_arg_type(&iter);
        if(type!=DBUS_TYPE_STRING)
        {
            DBG("Invalid notify message. Notrification type argument is not a string\n");
            RET(0);
        }
        dbus_message_iter_get_basic(&iter, &notification_type);
        DBG("notification_type; '%s'\n", notification_type);
        dbus_message_iter_next(&iter);
    
        // Urgency level -- not yet implempeted
        type=dbus_message_iter_get_arg_type(&iter);
        if(type!=DBUS_TYPE_BYTE)
        {
            DBG("Invalid notify message. Urgency argument is not a byte\n");
            RET(0);
        }
        dbus_message_iter_get_basic(&iter, &urgency_level);
        DBG("urgency_level; %d\n", urgency_level);
        dbus_message_iter_next(&iter);
    }

    // Summary
    type=dbus_message_iter_get_arg_type(&iter);
    if(type!=DBUS_TYPE_STRING)
    {
        DBG("Invalid notify message. Summary argument is not a string\n");
        RET(0);
    }
    dbus_message_iter_get_basic(&iter, &summary);
    DBG("title; '%s'\n", summary);
    dbus_message_iter_next(&iter);

    // Body
    type=dbus_message_iter_get_arg_type(&iter);
    if(type!=DBUS_TYPE_STRING)
    {
        DBG("Invalid notify message. Summary argument is not a string\n");
        RET(0);
    }
    dbus_message_iter_get_basic(&iter, &body);
    DBG("body; '%s'\n", body);
    dbus_message_iter_next(&iter);

    // Images
    type=dbus_message_iter_get_arg_type(&iter);
    if(type==DBUS_TYPE_STRING)
    {
        dbus_message_iter_get_basic(&iter, &images);
        DBG("app_icon; '%s'\n", images);
    }
    else if(type==DBUS_TYPE_ARRAY)
    {
        images=ib_dbus_extract_image_from_array(&iter);
    }
    else
    {
        DBG("Invalid notify message. images argument is not an array or a string\n");
        RET(0);
    }
    dbus_message_iter_next(&iter);

    if(shortmsg==FALSE)
    {
        // Actions -- we don't care
        dbus_message_iter_next(&iter);
    
        // Hints -- we don't care
        dbus_message_iter_next(&iter);
    
        // Expires
        type=dbus_message_iter_get_arg_type(&iter);
        if(type!=DBUS_TYPE_BOOLEAN)
        {
            DBG("Invalid notify message. Expires argument is not boolean\n");
            RET(0);
        }
        dbus_message_iter_get_basic(&iter, &expires);
        DBG("expires; %d\n", expires);
        dbus_message_iter_next(&iter);
    
        // Expire timeout
        type=dbus_message_iter_get_arg_type(&iter);
        if(type!=DBUS_TYPE_UINT32)
        {
            DBG("Invalid notify message. Expires argument is not uint32\n");
            RET(0);
        }
        dbus_message_iter_get_basic(&iter, &expire_timeout);
        DBG("expire_timeout; %d\n", expire_timeout);
        dbus_message_iter_next(&iter);
    }


    // Let's show our notify now... at last ;)
    if(app_icon && *app_icon)
    {
        GtkRequisition req;

        gtk_widget_size_request(ib.time, &req);
        icon=ib_gui_load_image(app_icon, FALSE, 0, req.height+2, req.height+2); //icon should be a little bigger than text
    }

    if(app_name && *app_name)
    {
        if(summary && *summary)
            title=g_strdup_printf("%s - %s", app_name, summary);
        else
            title=app_name;
    }
    else if(summary && *summary)
        title=summary;
    else
        title="";

    if(images && *images)
        pix=(GdkPixbufAnimation *) ib_gui_load_image(images, FALSE, 0, ib.imagewmax, ib.imagehmax);

    ib_gui_notify_show(icon, title, body, pix, expire_timeout);

    if(app_name && *app_name && !summary && !*summary)
        g_free(title);
    if(pix)
        g_object_unref(G_OBJECT(pix));

    ++ib.notifyid;

    if(shortmsg==TRUE)
        RET(NULL);
    reply=dbus_message_new_method_return(message);
    dbus_message_iter_init_append(reply, &iter);
    id=(replaces_id)?replaces_id:ib.notifyid;
    dbus_message_iter_append_basic(&iter, DBUS_TYPE_UINT32, &id);

    RET(reply);
}

static DBusMessage *ib_dbus_handle_close(DBusMessage *message)
{
    ENTER;
    guint id;

    if(!dbus_message_get_args(message, NULL, DBUS_TYPE_UINT32, &id, DBUS_TYPE_INVALID))
    {
        DBG("Error, wrong CloseNotification message\n");
        RET(NULL);
    }

    if(id==ib.notifyid && ib.notifyhided==FALSE)
        ib_gui_notify_hide(ib.aftertimeout, NOTIFY_CLOSENOTIFY);
    RET(NULL);
}

static DBusMessage *ib_dbus_handle_get_info(DBusMessage *message)
{
    ENTER;
    DBusMessage *reply;
    DBusMessageIter iter;
    const gchar *name="InfoBox";
    const gchar *vendor="freedesktop.org";
    const gchar *version=VERSION;

    reply=dbus_message_new_method_return(message);
    dbus_message_iter_init_append(reply, &iter);
    dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &name);
    dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &vendor);
    dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &version);

    RET(reply);
}

static DBusMessage *ib_dbus_handle_get_caps(DBusMessage *message)
{
    ENTER;
    DBusMessage *reply;
    DBusMessageIter iter, array_iter;
    register gint i;
    static const char *caps[]={"body", "body-hyperlinks", "body-markup", "icon-multi"};

    reply=dbus_message_new_method_return(message);
    dbus_message_iter_init_append(reply, &iter);
    dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY, DBUS_TYPE_STRING_AS_STRING, &array_iter);
    for(i=0; i<G_N_ELEMENTS(caps); i++)
        dbus_message_iter_append_basic(&array_iter, DBUS_TYPE_STRING, &caps[i]);
    dbus_message_iter_close_container(&iter, &array_iter);

    RET(reply);
}

static DBusMessage *ib_dbus_handle_set_config(DBusMessage *message)
{
    ENTER;
    GKeyFile *keyfile;
    GError *error=NULL;
    gchar *data, *startgroup, **keys;
    gint len;

    if(!dbus_message_get_args(message, NULL, DBUS_TYPE_STRING, &data, DBUS_TYPE_INVALID))
    {
        DBG("Error, wrong SetServerConfig message\n");
        RET(NULL);
    }

	keyfile=g_key_file_new();
	if(!g_key_file_load_from_data(keyfile, data, (gsize) strlen(data), G_KEY_FILE_NONE, NULL))
    {
        DBG("Couldn't read config data: %s\n", error->message);
        g_error_free(error);
        RET(NULL);
    }
	startgroup=g_key_file_get_start_group(keyfile);
    if(ib_config_strcmp(startgroup, "InfoBoxDaemon")!=0)
    {
        DBG("Error in config file: wrong group name, should be \"InfoBoxd\"\n");
        RET(NULL);
    }

	keys=g_key_file_get_keys(keyfile, startgroup, &len, NULL);
    ib_config_parser(keyfile, startgroup, keys);
    ib_config_setup_gui();

	g_strfreev(keys);
	g_free(startgroup);
	g_key_file_free(keyfile);

    RET(NULL);
}

static DBusHandlerResult ib_dbus_filter_func(DBusConnection *conn, DBusMessage *message, void *user_data)
{
    ENTER;
    int message_type;
    
    message_type=dbus_message_get_type(message);
    if(message_type==DBUS_MESSAGE_TYPE_ERROR)
    {
        DBG("Error received: %s\n", dbus_message_get_error_name(message));
        RET(DBUS_HANDLER_RESULT_HANDLED);
    }
    else if(message_type==DBUS_MESSAGE_TYPE_SIGNAL)
    {
		const gchar *member;

        member=dbus_message_get_member(message);
        if(member=="ServiceAcquired")
            RET(DBUS_HANDLER_RESULT_HANDLED);
        DBG("Received signal %s\n", member);
        RET(DBUS_HANDLER_RESULT_NOT_YET_HANDLED);
    }
    else
    {
        const gchar *member, *path, *iface;
        DBusMessage *ret=NULL;

        member=dbus_message_get_member(message);
        DBG("method=%s\n", member);

        path=dbus_message_get_path(message);
        if(g_ascii_strcasecmp(path, "/org/freedesktop/Notifications")!=0)
        {
            DBG("message received on unknown object '%s'\n", path);
            RET(DBUS_HANDLER_RESULT_NOT_YET_HANDLED);
        }

        iface=dbus_message_get_interface(message);
        if(g_ascii_strcasecmp(iface, "org.freedesktop.Notifications")!=0)
        {
            DBG("unknown message received: %s.%s\n", iface, member);
            RET(DBUS_HANDLER_RESULT_NOT_YET_HANDLED);
        }
    
        if(g_ascii_strcasecmp(member, "Notify")==0)
        {
            DBG("Notify\n");
            ret=ib_dbus_handle_notify(message, FALSE);
        }
        else if(g_ascii_strcasecmp(member, "ShortNotify")==0)
        {
            DBG("ShortNotify\n");
            ret=ib_dbus_handle_notify(message, TRUE);
        }
        else if(g_ascii_strcasecmp(member, "GetCapabilities")==0)
        {
            DBG("GetCapabilities\n");
            ret=ib_dbus_handle_get_caps(message);
        }
        else if(g_ascii_strcasecmp(member, "GetServerInformation")==0)
        {
            DBG("GetServerInformation\n");
            ret=ib_dbus_handle_get_info(message);
        }
        else if(g_ascii_strcasecmp(member, "CloseNotification")==0)
        {
            DBG("CloseNotification\n");
            ret=ib_dbus_handle_close(message);
        }
        else if(g_ascii_strcasecmp(member, "SetServerConfig")==0)
        {
            DBG("SetServerConfig\n");
            ret=ib_dbus_handle_set_config(message);
        }
        else
        {
            DBG("Unknown method: %s\n", member);
            RET(DBUS_HANDLER_RESULT_NOT_YET_HANDLED);
        }

        // we always reply to messages, even if it's just empty
        if(!ret)
            ret=dbus_message_new_method_return(message);
        dbus_connection_send(conn, ret, NULL);
        dbus_message_unref(ret);
        DBG("sent reply\n");

        RET(DBUS_HANDLER_RESULT_HANDLED);
    }
}

static void ib_dbus_init(void)
{
    ENTER;
    DBusError error;

    dbus_error_init(&error);

    ib.dbusconn=dbus_bus_get(DBUS_BUS_SESSION, &error);
    if(ib.dbusconn==NULL)
    {
        g_fprintf(stderr, "Unable to get session bus: %s, perhaps you "
                "need to start DBUS?\n", error.message);
        exit(1);
    }
    dbus_connection_setup_with_g_main(ib.dbusconn, NULL);

    dbus_bus_request_name(ib.dbusconn, "org.freedesktop.Notifications", 0, &error);
    if(dbus_error_is_set(&error))
    {
        g_fprintf(stderr, "Unable to acquire service org.freedesktop.Notifications: %s\n",
                error.message);
        exit(1);
    }
    dbus_connection_add_filter(ib.dbusconn, ib_dbus_filter_func, NULL, NULL);

    RET();
}
/*====================================================================*/


/**********************************************************************/
/*                            main control                            */
int main(int argc, char *argv[], char *env[])
{
    gtk_init(&argc, &argv);
	ib_config_set();
    ib_dbus_init();
	ib_gui_start();
    ib_config_setup_gui();
	gtk_main ();

	return(0);
}
/*====================================================================*/
